import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/scheduler.dart';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';
import 'package:adeptutils/adeptutils.dart';
import 'dart:ui' as ui show BoxHeightStyle, BoxWidthStyle;

enum CustomButton {
  NONE,
  ALWAYS,
  ON_TEXT,
}

enum CustomButtonDefaultAction {
  NONE,
  CLEAR,
  PASSWORD_TOGGLE,
}

class AdeptTextField extends StatefulWidget {
  final Key? key;
  final AdeptTextEditingController? controller;
  final FocusNode? focusNode;
  final InputDecoration decoration;
  final TextInputType? keyboardType;
  final TextInputAction? textInputAction;
  final TextCapitalization textCapitalization;
  final TextStyle? style;
  final TextAlign textAlign;
  final TextDirection? textDirection;
  final bool autofocus;
  final bool obscureText;
  final bool autocorrect;
  final int maxLines;
  final int? maxLength;
  final MaxLengthEnforcement? maxLengthEnforcement;
  final ValueChanged<String>? onChanged;
  final VoidCallback? onEditingComplete;
  final ValueChanged<String>? onSubmitted;
  final List<TextInputFormatter>? inputFormatters;
  final bool? enabled;
  final double cursorWidth;
  final Radius? cursorRadius;
  final Color? cursorColor;
  final Brightness? keyboardAppearance;
  final EdgeInsets scrollPadding;
  final bool enableInteractiveSelection;
  final TextSelectionControls? selectionControls;
  final GestureTapCallback? onTap;
  final StrutStyle? strutStyle;
  final TextAlignVertical? textAlignVertical;
  final bool readOnly;
  final ToolbarOptions? toolbarOptions;
  final bool? showCursor;
  final String obscuringCharacter;
  final SmartDashesType? smartDashesType;
  final SmartQuotesType? smartQuotesType;
  final bool enableSuggestions;
  final int? minLines;
  final bool expands;
  final AppPrivateCommandCallback? onAppPrivateCommand;
  final double? cursorHeight;
  final ui.BoxHeightStyle selectionHeightStyle;
  final ui.BoxWidthStyle selectionWidthStyle;
  final DragStartBehavior dragStartBehavior;
  final MouseCursor? mouseCursor;
  final InputCounterWidgetBuilder? buildCounter;
  final ScrollController? scrollController;
  final ScrollPhysics? scrollPhysics;
  final Iterable<String>? autofillHints;
  final String? restorationId;

  final CustomButton customButton;
  final CustomButtonDefaultAction customButtonDefaultAction;
  final VoidCallback? customButtonAction;
  final Color? customButtonColor;
  final IconData? customButtonIcon;
  final double customButtonIconSize;
  final double customButtonIconPadding;
  final bool clearErrorOnTextChange;

  AdeptTextField({
    this.key,
    this.controller,
    this.focusNode,
    InputDecoration? decoration,
    this.keyboardType,
    this.textInputAction,
    TextCapitalization? textCapitalization,
    this.style,
    TextAlign? textAlign,
    this.textDirection,
    bool? autofocus,
    bool? obscureText,
    bool? autocorrect,
    int? maxLines,
    this.maxLength,
    this.maxLengthEnforcement,
    this.onChanged,
    this.onEditingComplete,
    this.onSubmitted,
    this.inputFormatters,
    this.enabled,
    double? cursorWidth,
    this.cursorRadius,
    this.cursorColor,
    this.keyboardAppearance,
    EdgeInsets? scrollPadding,
    bool? enableInteractiveSelection,
    this.selectionControls,
    this.onTap,
    this.strutStyle,
    this.textAlignVertical,
    bool? readOnly,
    this.toolbarOptions,
    this.showCursor,
    String? obscuringCharacter,
    this.smartDashesType,
    this.smartQuotesType,
    bool? enableSuggestions,
    this.minLines,
    bool? expands,
    this.onAppPrivateCommand,
    this.cursorHeight,
    ui.BoxHeightStyle? selectionHeightStyle,
    ui.BoxWidthStyle? selectionWidthStyle,
    DragStartBehavior? dragStartBehavior,
    this.mouseCursor,
    this.buildCounter,
    this.scrollController,
    this.scrollPhysics,
    this.autofillHints,
    this.restorationId,
    CustomButton? customButton,
    CustomButtonDefaultAction? customButtonDefaultAction,
    this.customButtonAction,
    this.customButtonColor,
    this.customButtonIcon,
    double? customButtonIconSize,
    double? customButtonIconPadding,
    bool? clearErrorOnTextChange = true,
  })  : this.decoration = decoration ?? const InputDecoration(),
        this.textCapitalization = textCapitalization ?? TextCapitalization.none,
        this.textAlign = textAlign ?? TextAlign.start,
        this.autofocus = autofocus ?? false,
        this.obscureText = obscureText ?? false,
        this.autocorrect = autocorrect ?? true,
        this.maxLines = maxLines ?? 1,
        this.cursorWidth = cursorWidth ?? 2.0,
        this.scrollPadding = scrollPadding ?? const EdgeInsets.all(20.0),
        this.enableInteractiveSelection = enableInteractiveSelection ?? true,
        this.readOnly = readOnly ?? false,
        this.obscuringCharacter = obscuringCharacter ?? '•',
        this.enableSuggestions = enableSuggestions ?? true,
        this.expands = expands ?? false,
        this.selectionHeightStyle =
            selectionHeightStyle ?? ui.BoxHeightStyle.tight,
        this.selectionWidthStyle =
            selectionWidthStyle ?? ui.BoxWidthStyle.tight,
        this.dragStartBehavior = dragStartBehavior ?? DragStartBehavior.start,
        this.customButton = customButton ?? CustomButton.NONE,
        this.customButtonDefaultAction =
            customButtonDefaultAction ?? CustomButtonDefaultAction.NONE,
        this.customButtonIconSize = customButtonIconSize ?? 24,
        this.customButtonIconPadding = customButtonIconPadding ?? 8,
        this.clearErrorOnTextChange = clearErrorOnTextChange ?? true;

  @override
  _AdeptTextFieldState createState() => _AdeptTextFieldState();
}

class _AdeptTextFieldState extends State<AdeptTextField> {
  late final AdeptTextEditingController controller;
  bool hasText = false;
  bool hasError = false;
  String? lastText;
  late bool obscureText;
  late InputDecoration decoration;

  _AdeptTextFieldState();

  @override
  void initState() {
    super.initState();
    controller = widget.controller ?? AdeptTextEditingController();
    if (widget.customButtonDefaultAction ==
        CustomButtonDefaultAction.PASSWORD_TOGGLE)
      obscureText = true;
    else
      obscureText = widget.obscureText;
    decoration = widget.decoration;
    hasText = controllerHasText;
    hasError = controller.hasError;
    lastText = controller.text;
    controller.addListener(onTextChanged);
  }

  void updateDecoration() {
    if (widget.customButton != CustomButton.NONE) {
      decoration = decoration.copyWith(
        errorText: controller.error,
        suffixIcon: !hasText
            ? SizedBox()
            : IconButton(
                icon: getCustomButtonIcon,
                iconSize: widget.customButtonIconSize,
                padding: EdgeInsets.all(widget.customButtonIconPadding),
                onPressed: () {
                  if (widget.customButtonAction != null)
                    widget.customButtonAction?.call();
                  else {
                    switch (widget.customButtonDefaultAction) {
                      case CustomButtonDefaultAction.CLEAR:
                        controller.clear();
                        break;
                      case CustomButtonDefaultAction.PASSWORD_TOGGLE:
                        obscureText = !obscureText;
                        setState(() {});
                        break;
                      default:
                    }
                  }
                },
              ),
        suffixText: null,
      );
    }
  }

  bool get controllerHasText {
    bool show = false;
    if (widget.customButton == CustomButton.ALWAYS) show = true;
    if (widget.customButton == CustomButton.ON_TEXT &&
        controller.text.isNotEmpty) show = true;
    return show;
  }

  void onTextChanged() {
    bool controllerHasText = this.controllerHasText;
    if (controllerHasText != hasText) {
      hasText = controllerHasText;
      hasError = controller.hasError;
      setState(() {});
    } else if (hasError != controller.hasError) {
      hasError = controller.hasError;
      setState(() {});
    }
    if (widget.onChanged != null && lastText != controller.text)
      widget.onChanged!(lastText = controller.text);
    lastText = controller.text;
  }

  Widget get getCustomButtonIcon {
    if (widget.customButtonIcon != null)
      return Icon(
        widget.customButtonIcon,
        color: widget.customButtonColor,
      );
    switch (widget.customButtonDefaultAction) {
      case CustomButtonDefaultAction.CLEAR:
        return Icon(
          Icons.clear,
          color: widget.customButtonColor,
        );
      case CustomButtonDefaultAction.PASSWORD_TOGGLE:
        return Icon(
          obscureText ? Icons.visibility_off : Icons.visibility,
          color: widget.customButtonColor,
        );
      default:
        return SizedBox();
    }
  }

  @override
  void didUpdateWidget(covariant AdeptTextField oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (oldWidget.controller != controller) {
      oldWidget.controller?.removeListener(onTextChanged);
      controller.addListener(onTextChanged);
    }
  }

  @override
  Widget build(BuildContext context) {
    updateDecoration();
    return Theme(
      data: Theme.of(context).copyWith(splashColor: Colors.transparent),
      child: TextField(
        key: widget.key,
        controller: controller,
        focusNode: widget.focusNode,
        decoration: decoration,
        keyboardType: widget.keyboardType,
        textInputAction: widget.textInputAction,
        textCapitalization: widget.textCapitalization,
        style: widget.style,
        textAlign: widget.textAlign,
        textDirection: widget.textDirection,
        autofocus: widget.autofocus,
        obscureText: obscureText,
        autocorrect: widget.autocorrect,
        maxLines: widget.maxLines,
        maxLength: widget.maxLength,
        maxLengthEnforcement: widget.maxLengthEnforcement,
//        onChanged: widget.onChanged, //onChanged function is managed above in the onTextChanged function
        onEditingComplete: widget.onEditingComplete,
        onSubmitted: widget.onSubmitted,
        inputFormatters: widget.inputFormatters,
        enabled: widget.enabled,
        cursorWidth: widget.cursorWidth,
        cursorRadius: widget.cursorRadius,
        cursorColor: widget.cursorColor,
        keyboardAppearance: widget.keyboardAppearance,
        scrollPadding: widget.scrollPadding,
        enableInteractiveSelection: widget.enableInteractiveSelection,
        selectionControls: widget.selectionControls,
        onTap: widget.onTap,
        strutStyle: widget.strutStyle,
        textAlignVertical: widget.textAlignVertical,
        readOnly: widget.readOnly,
        toolbarOptions: widget.toolbarOptions,
        showCursor: widget.showCursor,
        obscuringCharacter: widget.obscuringCharacter,
        smartDashesType: widget.smartDashesType,
        smartQuotesType: widget.smartQuotesType,
        enableSuggestions: widget.enableSuggestions,
        minLines: widget.minLines,
        expands: widget.expands,
        onAppPrivateCommand: widget.onAppPrivateCommand,
        cursorHeight: widget.cursorHeight,
        selectionHeightStyle: widget.selectionHeightStyle,
        selectionWidthStyle: widget.selectionWidthStyle,
        dragStartBehavior: widget.dragStartBehavior,
        mouseCursor: widget.mouseCursor,
        buildCounter: widget.buildCounter,
        scrollController: widget.scrollController,
        scrollPhysics: widget.scrollPhysics,
        autofillHints: widget.autofillHints,
        restorationId: widget.restorationId,
      ),
    );
  }

  @override
  void dispose() {
    controller.removeListener(onTextChanged);
    super.dispose();
  }
}
